@@ -1,4 +1,4 @@
-{-# LANGUAGE FlexibleContexts, FlexibleInstances, GADTs #-}
+{-# LANGUAGE FlexibleContexts, FlexibleInstances, GADTs, RankNTypes #-}
 -- | The @esqueleto@ EDSL (embedded domain specific language).
 -- This module replaces @Database.Persist@, so instead of
 -- importing that module you should just import this one:
@@ -38,28 +38,34 @@ module Database.Esqueleto
     -- $gettingstarted
 
     -- * @esqueleto@'s Language
-    Esqueleto( where_, on, groupBy, orderBy, rand, asc, desc, limit, offset
+    where_, on, groupBy, orderBy, rand, asc, desc, limit, offset
              , distinct, distinctOn, don, distinctOnOrderBy, having, locking
-             , sub_select, sub_selectDistinct, (^.), (?.)
-             , val, isNothing, just, nothing, joinV
+             , sub_select, (^.), (?.)
+             , val, isNothing, just, nothing, joinV, withNonNull
              , countRows, count, countDistinct
              , not_, (==.), (>=.), (>.), (<=.), (<.), (!=.), (&&.), (||.)
-             , (+.), (-.), (/.), (*.)
+             , between, (+.), (-.), (/.), (*.)
              , random_, round_, ceiling_, floor_
              , min_, max_, sum_, avg_, castNum, castNumM
              , coalesce, coalesceDefault
-             , lower_, like, ilike, (%), concat_, (++.), castString
-             , subList_select, subList_selectDistinct, valList, justList
+             , lower_, upper_, trim_, ltrim_, rtrim_, length_, left_, right_
+             , like, ilike, (%), concat_, (++.), castString
+             , subList_select, valList, justList
              , in_, notIn, exists, notExists
              , set, (=.), (+=.), (-=.), (*=.), (/=.)
-             , case_, toBaseId)
+             , case_, toBaseId
+  , subSelect
+  , subSelectMaybe
+  , subSelectCount
+  , subSelectForeign
+  , subSelectList
+  , subSelectUnsafe
   , ToBaseId(..)
   , when_
   , then_
   , else_
   , from
   , Value(..)
-  , unValue
   , ValueList(..)
   , OrderBy
   , DistinctOn
@@ -71,32 +77,37 @@ module Database.Esqueleto
   , LeftOuterJoin(..)
   , RightOuterJoin(..)
   , FullOuterJoin(..)
+  , JoinKind(..)
   , OnClauseWithoutMatchingJoinException(..)
-
     -- * SQL backend
   , SqlQuery
   , SqlExpr
   , SqlEntity
   , select
-  , selectDistinct
   , selectSource
-  , selectDistinctSource
   , delete
   , deleteCount
   , update
   , updateCount
   , insertSelect
   , insertSelectCount
-  , insertSelectDistinct
   , (<#)
   , (<&>)
-
+  -- ** Rendering Queries
+  , renderQueryToText
+  , renderQuerySelect
+  , renderQueryUpdate
+  , renderQueryDelete
+  , renderQueryInsertInto
+    -- * Internal.Language
+  , From
     -- * RDBMS-specific modules
     -- $rdbmsSpecificModules
 
     -- * Helpers
   , valkey
   , valJ
+  , associateJoin
 
     -- * Re-exports
     -- $reexports
@@ -107,6 +118,7 @@ module Database.Esqueleto
 import Control.Monad.IO.Class (MonadIO)
 import Control.Monad.Trans.Reader (ReaderT)
 import Data.Int (Int64)
+import qualified Data.Map.Strict as Map
 import Database.Esqueleto.Internal.Language
 import Database.Esqueleto.Internal.Sql
 import Database.Esqueleto.Internal.PersistentImport
@@ -314,7 +326,7 @@ import qualified Database.Persist
 -- SELECT P1.*, Follow.*, P2.*
 -- FROM Person AS P1
 -- INNER JOIN Follow ON P1.id = Follow.follower
--- INNER JOIN P2 ON P2.id = Follow.followed
+-- INNER JOIN Person AS P2 ON P2.id = Follow.followed
 -- @
 --
 -- In @esqueleto@, we may write the same query above as:
@@ -322,16 +334,11 @@ import qualified Database.Persist
 -- @
 -- 'select' $
 -- 'from' $ \\(p1 `'InnerJoin`` f `'InnerJoin`` p2) -> do
--- 'on' (p2 '^.' PersonId '==.' f '^.' FollowFollowed)
 -- 'on' (p1 '^.' PersonId '==.' f '^.' FollowFollower)
+-- 'on' (p2 '^.' PersonId '==.' f '^.' FollowFollowed)
 -- return (p1, f, p2)
 -- @
 --
--- /Note carefully that the order of the ON clauses is/
--- /reversed!/ You're required to write your 'on's in reverse
--- order because that helps composability (see the documentation
--- of 'on' for more details).
---
 -- We also currently support @UPDATE@ and @DELETE@ statements.
 -- For example:
 --
@@ -405,8 +412,8 @@ import qualified Database.Persist
 
 -- | @valkey i = 'val' . 'toSqlKey'@
 -- (<https://github.com/prowdsponsor/esqueleto/issues/9>).
-valkey :: (Esqueleto query expr backend, ToBackendKey SqlBackend entity, PersistField (Key entity)) =>
-          Int64 -> expr (Value (Key entity))
+valkey :: (ToBackendKey SqlBackend entity, PersistField (Key entity)) =>
+          Int64 -> SqlExpr (Value (Key entity))
 valkey = val . toSqlKey
 
 
@@ -420,8 +427,8 @@ valkey = val . toSqlKey
 -- (<https://github.com/prowdsponsor/esqueleto/pull/69>).
 --
 -- /Since: 1.4.2/
-valJ :: (Esqueleto query expr backend, PersistField (Key entity)) =>
-        Value (Key entity) -> expr (Value (Key entity))
+valJ :: (PersistField (Key entity)) =>
+        Value (Key entity) -> SqlExpr (Value (Key entity))
 valJ = val . unValue
 
 
@@ -436,3 +443,28 @@ deleteKey :: ( PersistStore backend
              , PersistEntity val )
           => Key val -> ReaderT backend m ()
 deleteKey = Database.Persist.delete
+
+-- | Avoid N+1 queries and join entities into a map structure
+-- @
+-- getFoosAndNestedBarsFromParent :: ParentId -> (Map (Key Foo) (Foo, [Maybe (Entity Bar)]))
+-- getFoosAndNestedBarsFromParent parentId = 'fmap' associateJoin $ 'select' $
+-- 'from' $ \\(foo `'LeftOuterJoin`` bar) -> do
+--   'on' (bar '?.' BarFooId '==.' foo '^.' FooId)
+--   'where_' (foo '^.' FooParentId '==.' 'val' parentId)
+--   'pure' (foo, bar)
+-- @
+--
+-- @since 3.1.2
+associateJoin
+  :: forall e1 e0
+   . Ord (Key e0)
+  => [(Entity e0, e1)]
+  -> Map.Map (Key e0) (e0, [e1])
+associateJoin = foldr f start
+  where
+    start = Map.empty
+    f (one, many) =
+      Map.insertWith
+        (\(oneOld, manyOld) (_, manyNew) -> (oneOld, manyNew ++ manyOld ))
+        (entityKey one)
+        (entityVal one, [many])
