@@ -7,9 +7,15 @@
            , ScopedTypeVariables
            , TypeApplications
            , TypeFamilies
+           , PartialTypeSignatures
  #-}
 module Main (main) where
 
+import Data.Coerce
+import Data.Foldable
+import qualified Data.Map.Strict as Map
+import Data.Map (Map)
+import Data.Time
 import Control.Arrow ((&&&))
 import Control.Monad (void, when)
 import Control.Monad.Catch (MonadCatch, catch)
@@ -36,6 +42,7 @@ import Database.Persist.Postgresql (withPostgresqlConn)
 import Database.PostgreSQL.Simple (SqlError(..), ExecStatus(..))
 import System.Environment
 import Test.Hspec
+import Test.Hspec.QuickCheck
 
 import Common.Test
 import PostgreSQL.MigrateJSON
@@ -52,10 +59,6 @@ testPostgresqlCoalesce = do
         return (coalesce [p ^. PersonAge])
       return ()
 
-
-
-
-
 nameContains :: (BaseBackend backend ~ SqlBackend,
                  BackendCompatible SqlBackend backend,
                  MonadIO m, SqlString s,
@@ -486,6 +489,55 @@ testAggregateFunctions = do
 
 testPostgresModule :: Spec
 testPostgresModule = do
+  describe "date_trunc" $ modifyMaxSuccess (`div` 10) $ do
+    prop "works" $ \listOfDateParts -> run $ do
+      let
+        utcTimes =
+          map
+            (\(y, m, d, s) ->
+              fromInteger s
+                `addUTCTime`
+                  UTCTime (fromGregorian (2000 + y) m d) 0
+            )
+          listOfDateParts
+        truncateDate
+          :: SqlExpr (Value String)  -- ^ .e.g (val "day")
+          -> SqlExpr (Value UTCTime) -- ^ input field
+          -> SqlExpr (Value UTCTime) -- ^ truncated date
+        truncateDate datePart expr =
+          ES.unsafeSqlFunction "date_trunc" (datePart, expr)
+        vals =
+          zip (map (DateTruncTestKey . fromInteger) [1..]) utcTimes
+      for_ vals $ \(idx, utcTime) -> do
+        insertKey idx (DateTruncTest utcTime)
+
+      -- Necessary to get the test to pass; see the discussion in
+      -- https://github.com/bitemyapp/esqueleto/pull/180
+      rawExecute "SET TIME ZONE 'UTC'" []
+      ret <-
+        fmap (Map.fromList . coerce :: _ -> Map DateTruncTestId (UTCTime, UTCTime)) $
+        select $
+        from $ \dt -> do
+        pure
+          ( dt ^. DateTruncTestId
+          , ( dt ^. DateTruncTestCreated
+            , truncateDate (val "day") (dt ^. DateTruncTestCreated)
+            )
+          )
+
+      liftIO $ for_ vals $ \(idx, utcTime) -> do
+        case Map.lookup idx ret of
+          Nothing ->
+            expectationFailure "index not found"
+          Just (original, truncated) -> do
+            utcTime `shouldBe` original
+            if utctDay utcTime == utctDay truncated
+              then
+                utctDay utcTime `shouldBe` utctDay truncated
+              else
+                -- use this if/else to get a better error message
+                utcTime `shouldBe` truncated
+
   describe "PostgreSQL module" $ do
     describe "Aggregate functions" testAggregateFunctions
     it "chr looks sane" $
